import os, textwrap
import numpy as np
import unittest
from context import dpti
# from numpy.testing import assert_almost_equal
from dpti.einstein import frenkel, free_energy, ideal_gas_fe
lambda_seq = [
    "0.00:0.05:0.010",
    "0.05:0.15:0.020",
    "0.15:0.35:0.040",
    "0.35:1.00:0.065",
    "1"
]

class TestEinstein(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_frenkel(self):
        fe1 = -0.14061204010964043
        fe2 = frenkel('hti_test_files/frenkel')
        self.assertAlmostEqual(fe1, fe2)

    def test_vega(self):
        fe1 = -0.13882760104909486
        fe2 = free_energy('hti_test_files/vega')
        self.assertAlmostEqual(fe1, fe2)

    def test_ideal(self):
        fe1 = -1.8983591660560315
        fe2 = ideal_gas_fe('hti_test_files/ideal')
        # print('ideal_gas fe', fe2)
        self.assertAlmostEqual(fe1, fe2)

        # print(ideal_gas_fe)
        # pass
        # print(fe2)
        # self.assertAlmostEqual(fe1, fe2)
        # assert_almost_equal(array1, array2, decimal=10)

    # def test_protect_eps(self):
    #     array1 = parse_seq(lambda_seq, protect_eps=1e-6)
    #     assert_almost_equal(array1, array2, decimal=10)

    # def test_no_posi_args(self):
    #     with self.assertRaises(TypeError):
    #         parse_seq(lambda_seq, 1e-6)
            # assert_almost_equal(array1, array2, decimal=10)

# class TestBlockAvg(unittest.TestCase):
#     def setUp(self):
#         self.maxDiff = None

#     def test_normal(self):
#         avg1 = 7.158014302
#         err1 = 0.262593991
#         data_file = 'lammps_test_files/get_thermo.data'
#         data_array = np.loadtxt(data_file)
#         avg2, err2 = block_avg(data_array[:,1])
#         self.assertAlmostEqual(avg1, avg2, places=8)
#         self.assertAlmostEqual(err1, err2, places=8)

# class TestIntegrateRangeHti(unittest.TestCase):
#     def setUp(self):
#         self.maxDiff = None
#         # lamb

#     def test_lamb_array_odd(self):
#         result1 = -0.07001571298782591
#         stt_err1 = 2.1394708051996743e-05
#         sys_err2 = 2.0797767427780528e-07
#         data = np.loadtxt('hti_test_files/odd.hti.out')
#         lamb_array = data[:,0]
#         dU_array = data[:,1]
#         dU_err_array = data[:,2]
#         result2, stt_err2, sys_err2 = integrate_range_hti(lamb_array, dU_array, dU_err_array)
#         self.assertAlmostEqual(result1, result2, places=8)
#         self.assertAlmostEqual(stt_err1, stt_err2, places=8)
#         self.assertAlmostEqual(sys_err2, sys_err2, places=8)

#     def test_lamb_array_even(self):
#         result1 = -35.48046669098458
#         stt_err1 = 0.0001625198805022198
#         sys_err2 = 8.812949063852216e-07
#         data = np.loadtxt('hti_test_files/even.hti.out')
#         lamb_array = data[:,0]
#         dU_array = data[:,1]
#         dU_err_array = data[:,2]
#         result2, stt_err2, sys_err2 = integrate_range_hti(lamb_array, dU_array, dU_err_array)
#         self.assertAlmostEqual(result1, result2, places=8)
#         self.assertAlmostEqual(stt_err1, stt_err2, places=8)
#         self.assertAlmostEqual(sys_err2, sys_err2, places=8)






if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from operator import sub
import os, sys, json, argparse, glob, shutil, time
from typing import Optional
import numpy as np
import scipy.constants as pc

from scipy.integrate import solve_ivp
# sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../'))
from dpti.lib.utils import create_path, relative_link_file
from dpti.lib.utils import get_first_matched_key_from_dict
from dpti.lib.utils import block_avg
from dpti.lib.lammps import get_natoms
from dpti.ti import _gen_lammps_input
from dpti import ti
import dargs
from dargs import dargs, Argument, Variant

# from dpti.workflow
# from lib.RemoteJob import SSHSession, JobStatus, SlurmJob, PBSJob
# from dpgen.dispatcher.Dispatcher import Dispatcher
from dpdispatcher import Submission, Task, Resources, Machine

# try:
#     from airflow.exceptions import AirflowFailException, AirflowSkipException, DagNotFound, DagRunAlreadyExists
#     from airflow.api.client.local_client import Client
# except ImportError:
#     pass

# def _group_slurm_jobs(ssh_sess,
#                       resources,
#                       command,
#                       work_path,
#                       tasks,
#                       group_size,
#                       forward_common_files,
#                       forward_task_files,
#                       backward_task_files,
#                       remote_job = None) :
#     task_chunks = [
#         [os.path.basename(j) for j in tasks[i:i + group_size]] \
#         for i in range(0, len(tasks), group_size)
#     ]
#     job_list = []
#     for chunk in task_chunks :
#         rjob = remote_job(ssh_sess, work_path)
#         rjob.upload('.',  forward_common_files)
#         rjob.upload(chunk, forward_task_files)
#         rjob.submit(chunk, command, resources = resources)
#         job_list.append(rjob)

#     job_fin = [False for ii in job_list]
#     while not all(job_fin) :
#         for idx,rjob in enumerate(job_list) :
#             if not job_fin[idx] :
#                 status = rjob.check_status()
#                 if status == JobStatus.terminated :
#                     raise RuntimeError("find unsuccessfully terminated job in %s" % rjob.get_job_root())
#                 elif status == JobStatus.finished :
#                     rjob.download(task_chunks[idx], backward_task_files)
#                     rjob.clean()
#                     job_fin[idx] = True
#         time.sleep(30)

def _make_tasks_onephase(temp, 
                         pres, 
                         task_path, 
                         jdata,
                         ens = 'npt',
                         conf_file = 'conf.lmp', 
                         graph_file = 'graph.pb',
                         if_meam=False,
                         meam_model=None):
    # assume that model and conf.lmp exist in the current dir
    assert os.path.isfile(conf_file), (conf_file, os.getcwd())
    # assert(os.path.isfile(graph_file))
    conf_file = os.path.abspath(conf_file)
    if graph_file:
        graph_abs_file = os.path.abspath(graph_file)
    
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['mass_map', 'model_mass_map'])
    # MD simulation protocol
    nsteps = jdata['nsteps']
    # timestep = jdata['timestep']
    timestep = get_first_matched_key_from_dict(jdata, ['timestep', 'dt'])
    # thermo_freq = jdata['thermo_freq']
    thermo_freq = get_first_matched_key_from_dict(jdata, ['thermo_freq', 'stat_freq'])
    tau_t = jdata['tau_t']
    tau_p = jdata['tau_p']

    cwd = os.getcwd()
    if not os.path.isdir(task_path):
        create_path(task_path)

    os.chdir(task_path)
    if not os.path.exists('conf.lmp'):
        os.symlink(os.path.relpath(conf_file), 'conf.lmp')
    if graph_file:
        if not os.path.exists('graph.pb'):
            os.symlink(os.path.relpath(graph_abs_file), 'graph.pb')

    if if_meam:
        relative_link_file(meam_model['library_abs_path'], './')
        relative_link_file(meam_model['potential_abs_path'], './')
        # meam_library_basename = os.path.basename(meam_model['library'])
        # meam_potential_basename = os.path.basename(meam_model['potential'])
        # os.symlink(os.path.join('../../../', meam_library_basename), meam_library_basename)
        # os.symlink(os.path.join('../../../', meam_potential_basename), meam_potential_basename)
    
    # input for NPT MD
    lmp_str \
        = _gen_lammps_input('conf.lmp',
                               mass_map, 
                               graph_file,
                               nsteps, 
                               timestep,
                               ens,
                               temp,
                               pres,
                               tau_t = tau_t,
                               tau_p = tau_p,
                               thermo_freq = thermo_freq,
                               if_meam=if_meam,
                               meam_model=meam_model)
    with open('thermo.out', 'w') as fp :
        fp.write('%.16e %.16e' % (temp, pres))
    with open('in.lammps', 'w') as fp :
        fp.write(lmp_str)

    os.chdir(cwd)
    # end _make_tasks_onephase


def _setup_dpdt (task_path, jdata) :
    name_0 = jdata['phase_i']['name']
    name_1 = jdata['phase_ii']['name']
    conf_0 = os.path.join(task_path, '../', jdata['phase_i']['equi_conf'])
    conf_1 = os.path.join(task_path, '../', jdata['phase_ii']['equi_conf'])
    conf_0 = os.path.abspath(conf_0)
    conf_1 = os.path.abspath(conf_1)
    model = os.path.join(task_path, '../', jdata['model'])
    if model:
        model = os.path.abspath(model)

    task_abs_dir = create_path(task_path)
    conf_0_name = 'conf.%s.lmp' % '0'
    conf_1_name = 'conf.%s.lmp' % '1'
    # conf_0_name = 'conf.%s.lmp' % name_0
    # conf_1_name = 'conf.%s.lmp' % name_1
    copied_conf_0 = os.path.join(os.path.abspath(task_path), conf_0_name)
    copied_conf_1 = os.path.join(os.path.abspath(task_path), conf_1_name)
    shutil.copyfile(conf_0, copied_conf_0)
    shutil.copyfile(conf_1, copied_conf_1)
    if model:
        linked_model = os.path.join(os.path.abspath(task_path), 'graph.pb')
        shutil.copyfile(model, linked_model)

    with open(os.path.join(os.path.abspath(task_path), 'in.json'), 'w') as fp:
        json.dump(jdata, fp, indent=4)


def make_dpdt (temp,
               pres,
               inte_dir,
               task_path,
               mdata,
               natoms=None,
               shift=[0, 0],
               verbose=False,
               if_meam=False,
               meam_model=None,
               workflow=None):
    assert(os.path.isdir(task_path))    

    if if_meam:
        meam_model['library_abs_path'] = os.path.abspath(meam_model['library'])
        meam_model['potential_abs_path'] = os.path.abspath(meam_model['potential'])
        # relative_link_file(), task_path)
        # relative_link_file(os.path.abspath(meam_model['potential']), task_path)

    cwd = os.getcwd()
    os.chdir(task_path)

    # check if we need new MD simulations
    new_task = True
    if (not os.path.isdir('database')) or \
       (not os.path.isfile('database/dpdt.out')):
        if verbose :
            print('# dpdt: cannot find any MD record, start from scrtach')
        new_task = True
        counter = 0
    else :
        if verbose :
            print('# dpdt: found MD records, search if any record matches')
        data = np.loadtxt('database/dpdt.out')
        data = np.reshape(data, [-1,4])
        counter = data.shape[0]
        for ii in range(data.shape[0]) :
            if (np.linalg.norm(temp - data[ii][0]) < 1e-4) and \
               (np.linalg.norm(pres - data[ii][1]) < 1e-2) :
                if verbose :
                    print('# dpdt: found matched record at %f %f ' % (temp, pres))
                new_task = False
                dv = data[ii][2]
                dh = data[ii][3]
                break

    # try to find nearest simulation
    if new_task and os.path.isfile('database/dpdt.out'):
        data = np.loadtxt('database/dpdt.out')
        data = np.reshape(data, [-1,4])
        min_idx = -1
        min_val = 1e10
        if inte_dir == 't' :
            for ii in range(data.shape[0]) :
                dist = np.abs(data[ii][0] - temp)
                if dist < min_val :
                    min_val = dist
                    min_idx = ii
        elif inte_dir == 'p' :
            for ii in range(data.shape[0]) :
                dist = np.abs(data[ii][1] - pres)
                if dist < min_val :
                    min_val = dist
                    min_idx = ii
        else :
            raise RuntimeError("invalid inte_dir " + inte_dir)
        assert(min_idx >= 0)
        conf_0 = os.path.join('database', 'task.%06d' % min_idx, '0', 'out.lmp')
        conf_1 = os.path.join('database', 'task.%06d' % min_idx, '1', 'out.lmp')
    else :
        conf_0 = 'conf.0.lmp'
        conf_1 = 'conf.1.lmp'

    # new MD simulations are needed
    if new_task :
        if verbose :
            print('# dpdt: do not find any matched record, run new task from %d ' % counter)

        with open('in.json', 'r') as j:
            jdata = json.load(j)
        # make new task
        work_base = os.path.abspath(os.path.join('database', 'task.%06d' % counter))

        _make_tasks_onephase(temp, pres, 
                             os.path.join(work_base, '0'),
                             jdata,
                             ens = jdata['phase_i'].get('ens', None),
                             conf_file = conf_0,
                             graph_file = 'graph.pb',
                             if_meam=if_meam,
                             meam_model=meam_model)
        _make_tasks_onephase(temp, pres, 
                             os.path.join(work_base, '1'),
                             jdata, 
                             ens = jdata['phase_ii'].get('ens', None),
                             conf_file = conf_1,
                             graph_file = 'graph.pb',
                             if_meam=if_meam,
                             meam_model=meam_model)
        # submit new task

        # if workflow is None:
        machine = Machine.load_from_dict(mdata['machine'])
        resources = Resources.load_from_dict(mdata['resources'])

        command = 'lmp -i in.lammps'
        forward_files = ['conf.lmp', 'in.lammps', 'graph.pb']
        if if_meam:
            meam_library_basename = os.path.basename(meam_model['library'])
            meam_potential_basename = os.path.basename(meam_model['potential'])
            forward_files.extend([meam_library_basename, meam_potential_basename])
        backward_files = ['log.lammps', 'out.lmp']

        task_list = []
        for ii in range(2):
            task = Task(
                command=command,
                task_work_path=f'{ii}/',
                forward_files=forward_files,
                backward_files=backward_files
            )
            task_list.append(task)

        submission = Submission(
            work_base=work_base,
            machine=machine,
            resources=resources,
            forward_common_files=[],
            backward_common_files=[],
        )
        if workflow is None:
            submission.register_task_list(task_list=task_list)
            # submission.generate_jobs()
            submission.run_submission()
        else:
        # client.trigg
            workflow.trigger_loop(submission=submission, task_list=task_list, mdata=mdata)

        # run_tasks = ['0', '1']        

        # dispatcher.run_jobs(resources,
        #                     command,
        #                     work_path,
        #                     run_tasks,
        #                     1,
        #                     [],
        #                     forward_files,
        #                     backward_files,
        #                     forward_task_deference = False)
        # _group_slurm_jobs(ssh_sess,
        #                   resources,
        #                   command,
        #                   work_path,
        #                   run_tasks,
        #                   1,
        #                   [],
        #                   forward_files,
        #                   backward_files)

        # collect resutls
        log_0 = os.path.join(work_base, '0', 'log.lammps')
        log_1 = os.path.join(work_base, '1', 'log.lammps')
        if natoms == None :
            natoms = [get_natoms('conf.0.lmp'), get_natoms('conf.1.lmp')]
        stat_skip = jdata['stat_skip']
        stat_bsize = jdata['stat_bsize']
        t0 = ti._compute_thermo(log_0, natoms[0], stat_skip, stat_bsize)
        t1 = ti._compute_thermo(log_1, natoms[1], stat_skip, stat_bsize)
        dv = t1['v'] - t0['v']
        dh = t1['h'] - t0['h'] - (shift[1] - shift[0])
        with open(os.path.join('database', 'dpdt.out'), 'a') as fp:
            fp.write('%.16e %.16e %.16e %.16e\n' % \
                    (temp, pres, dv, dh))            
    os.chdir(cwd)
    return [dv, dh]


class GibbsDuhemFunc (object):
    def __init__ (self,
                  jdata,
                  mdata,
                  task_path,
                  inte_dir,
                  pref = 1.0,
                  natoms = None,
                  shift = [0, 0],
                  verbose = False,
                  if_meam=False,
                  meam_model=None,
                  workflow=None
    ):
        self.jdata = jdata
        self.mdata = mdata
        self.task_path = task_path
        self.inte_dir = inte_dir
        self.natoms = natoms
        self.verbose =  verbose
        self.pref = pref
        self.shift = shift
        self.if_meam = if_meam
        self.meam_model = meam_model
        self.workflow = workflow

        # self.dispatcher = Dispatcher(mdata['machine'], context_type = 'lazy-local', batch_type = 'pbs')
        if os.path.isdir(task_path) :
            print('find path ' + task_path + ' use it. The user should guarantee the consistency between the jdata and the found work path ')
        else :
            _setup_dpdt(task_path, jdata)

        self.ev2bar = pc.electron_volt / (pc.angstrom ** 3) * 1e-5

    def __call__ (self, x, y) :
        print('__call__', x, y)
        if self.inte_dir == 't' :
            # x: temp, y: pres
            [dv, dh] = make_dpdt(x, y,
                                 self.inte_dir,
                                 self.task_path, self.mdata,
                                 self.natoms,
                                 self.shift,
                                 self.verbose,
                                 if_meam=self.if_meam,
                                 meam_model=self.meam_model,
                                 workflow=self.workflow)
            return [dh / (x * dv) * self.ev2bar * self.pref]

        elif self.inte_dir == 'p' :
            # x: pres, y: temp
            [dv, dh] = make_dpdt(y, x,
                                 self.inte_dir,
                                 self.task_path, self.mdata,
                                 self.natoms,
                                 self.shift,
                                 self.verbose,
                                 if_meam=self.if_meam,
                                 meam_model=self.meam_model,
                                 workflow=self.workflow)
            return [(y * dv) / dh / self.ev2bar * (1/self.pref)]

# def gdi_main_loop(jdata, mdata, gdidata, begin=None, end=None, direction=None,
#     initial_value=None, step_value=None, abs_tol=10, rel_tol=0.01, if_water=None,
#     output=None, first_step=None, shift=[0.0, 0.0], verbose=None, if_meam=None, workflow=None):

def gdi_main_loop(jdata, mdata, gdidata_dict={}, gdidata_cli={}, workflow=None):

    gdiargs = [
        Argument("begin", float, optional=False),
        Argument("end", float, optional=False),
        Argument("direction", str, optional=False),
        Argument("initial_value", float, optional=False),
        Argument("step_value", [list, float], optional=True, default=None),
        Argument("abs_tol", float, optional=True, default=10,),
        Argument("rel_tol", float, optional=True, default=0.01),
        Argument("if_water", bool, optional=True, default=None),
        Argument("output", str, optional=True, default="new_job/"),
        Argument("first_step", float, optional=True, default=None),
        Argument("shift", list, optional=True, default=[0.0, 0.0]),
        Argument("verbose", bool, optional=True, default=True),
        Argument("if_meam", bool, optional=True, default=None),
    ]

    gdidata_format = Argument("gdidata", dict, gdiargs)

    gdidata = gdidata_format.normalize_value(gdidata_cli)

    if gdidata_dict:
        # gdidata = gdidata_dict.copy()
        gdidata = gdidata_format.normalize_value(gdidata_dict)
    print('debug', gdidata)
    # raise RuntimeError('gdidata')

    with open(os.path.join(os.path.dirname(os.path.abspath(gdidata['output'])), 
            'gdidata.run.json'), 'w') as f:
        json.dump(gdidata, f, indent=4)

    natoms = None
    if gdidata['if_water']:
        conf_0 = jdata['phase_i']['equi_conf']
        conf_1 = jdata['phase_ii']['equi_conf']
        natoms = [get_natoms(conf_0), get_natoms(conf_1)]
        natoms = [ii // 3  for ii in natoms]
    print ('# natoms: ', natoms)
    print ('# shifts: ', gdidata['shift'])
    meam_model = jdata.get('meam_model', None)

    # with open('')

    gdf = GibbsDuhemFunc(jdata,
                        mdata,
                        task_path=gdidata['output'],
                        inte_dir=gdidata['direction'],
                        natoms = natoms,
                        shift = gdidata['shift'],
                        verbose = gdidata['verbose'],
                        if_meam=gdidata['if_meam'],
                        meam_model=meam_model,
                        workflow=workflow)

    sol = solve_ivp(gdf,
                    [gdidata['begin'], gdidata['end']],
                    [gdidata['initial_value']],
                    t_eval = gdidata['step_value'],
                    method = 'RK23',
                    atol=gdidata['abs_tol'],
                    rtol=gdidata['rel_tol'],
                    first_step = gdidata['first_step']
                    )

    if gdidata['direction'] == 't' :
        tmp = np.concatenate([sol.t, sol.y[0]])
    else :
        tmp = np.concatenate([sol.y[0], sol.t])        

    tmp = np.reshape(tmp, [2,-1])
    np.savetxt(os.path.join(gdidata['output'], 'pb.out'), tmp.T)
    return True

def _main () :
    parser = argparse.ArgumentParser(
        description="Compute the phase boundary via Gibbs-Duhem integration")
    parser.add_argument('PARAM', type=str,
                        help='json parameter file')
    parser.add_argument('MACHINE', type=str,
                        help='json machine file')
    parser.add_argument('-g', '--gdidata-json', type=str, default=None,
                        help='json gdi integration file')
    parser.add_argument('-b','--begin', type=float, default=None,
                        help='start of the integration')
    parser.add_argument('-e','--end', type=float, default=None,
                        help='end of the integration')
    parser.add_argument('-d','--direction', type=str, choices=['t','p'], default=None,
                        help='direction of the integration, along T or P')
    parser.add_argument('-i','--initial-value', type=float, default=None,
                        help='the initial value of T (direction=p) or P (direction=t)')
    parser.add_argument('-s','--step-value', type=float, nargs = '+',
                        help='the T (direction=t) or P (direction=p) values must be evaluated')
    parser.add_argument('-a','--abs-tol', type=float, default = 10,
                        help='the absolute tolerance of the integration')
    parser.add_argument('-r','--rel-tol', type=float, default = 1e-2,
                        help='the relative tolerance of the integration')
    parser.add_argument('-w','--if-water', action = 'store_true',
                        help='assumes water molecules: nmols = natoms//3')
    parser.add_argument('-o','--output', type=str, default = 'new_job',
                        help='the output folder for the job')
    parser.add_argument('-f','--first-step', type=float, default=None,
                        help='the first step size of the integrator')
    parser.add_argument('-S','--shift', type=float, nargs = 2, default = [0.0, 0.0],
                        help='the output folder for the job')
    parser.add_argument('-v','--verbose', action = 'store_true',
                        help='print detailed infomation')
    parser.add_argument("-z", "--if-meam", help="whether use meam instead of dp", action="store_true")
    args = parser.parse_args()

    with open(args.PARAM) as j:
        jdata = json.load(j)
    with open(args.MACHINE) as m:
        mdata = json.load(m)

    if args.gdidata_json:
        with open(args.gdidata_json) as g:
            gdidata_dict = json.load(g)
    else:
        gdidata_dict=None

    gdidata_cli = dict(begin=args.begin, end=args.end,
        direction=args.direction, initial_value=args.initial_value, step_value=args.step_value, 
        abs_tol=args.abs_tol, rel_tol=args.rel_tol, if_water=args.if_water, output=args.output,
        first_step=args.first_step, shift=args.shift, verbose=args.verbose, if_meam=args.if_meam
    )

    return_value = gdi_main_loop(jdata=jdata, 
        mdata=mdata,
        gdidata_dict=gdidata_dict,
        gdidata_cli=gdidata_cli,
        workflow=None)
    # gdidata_run = gdiargs



    # return_value = gdi_main_loop(jdata=jdata, mdata=mdata, gdidata=gdidata, begin=args.begin, end=args.end,
    #     direction=args.direction, initial_value=args.initial_value, step_value=args.step_value, 
    #     abs_tol=args.abs_tol, rel_tol=args.rel_tol, if_water=args.if_water, output=args.output,
    #     first_step=args.first_step, shift=args.shift, verbose=args.verbose, if_meam=args.if_meam, workflow=None)

    return return_value

if __name__ == '__main__' :
    _main()        
    
# mdata = json.load(open('machine.json'))
# jdata = json.load(open('in.json'))
# # ssh_sess = SSHSession(mdata['machine'])
# # if not os.path.isdir('gdi_test') :
# #     _setup_dpdt('gdi_test', jdata)
# # make_dpdt(100, 1,  'gdi_test', mdata, ssh_sess, natoms = [96, 128])
# # make_dpdt(100, 20, 'gdi_test', mdata, ssh_sess, natoms = [96, 128])

# gdf = GibbsDuhemFunc(jdata, mdata, 'gdi_test', 'p', natoms = [96, 128], verbose = True)

# sol = solve_ivp(gdf, [1, 20000], [363], method = 'RK23', atol=10, rtol=1e-2)
# print(sol.t)
# print(sol.y)
# np.savetxt('t.out', sol.t)
# np.savetxt('p.out', sol.y)
# # print(gdf(100, 1))
# # print(gdf(100, 1))
# # print(gdf(200, 20))



#!/usr/bin/env python3

import numpy as np
from lib.utils import integrate
from lib.utils import integrate_sys_err
from lib.lammps import get_thermo
from lib.dump import split_traj
from lib.dump import get_posi
from lib.dump import get_atype
import lib.water as water
import lib.dump as dump

# def func (xx) :
#     return 0.02*xx*xx+0.01*xx+0.03

# x0 = np.arange(0,10.1)
# x1 = np.arange(0,10.1, 0.5)
# x2 = np.arange(0,10.1, 0.25)


# i0 = integrate(x0, func(x0), np.zeros(x0.shape))
# i1 = integrate(x1, func(x1), np.zeros(x1.shape))
# i2 = integrate(x2, func(x2), np.zeros(x2.shape))
# e0 = integrate_sys_err(x0, func(x0))
# e1 = integrate_sys_err(x1, func(x1))
# e2 = integrate_sys_err(x2, func(x2))

# print(i0[0], e0)
# print(i1[0], e1)
# print(i2[0], e2)

# get_thermo('log.lammps')

lines = open('dump.hti').read().split('\n')
ret = split_traj(lines) 
# print(get_posi(ret[0]))
# print(get_posi(ret[0])[127:130])
# print(get_atype(ret[0]))
# print(get_atype(ret[0])[127:130])

bd, tl = dump.get_dumpbox(ret[0])
orig, box = dump.dumpbox2box(bd, tl)
atype = dump.get_atype(ret[0])
posi = dump.get_posi(ret[0])
oh_list = water.min_oh_list(box, atype, posi)

for idx, ii in enumerate(ret) :
    bd, tl = dump.get_dumpbox(ii)
    orig, box = dump.dumpbox2box(bd, tl)
    posi = dump.get_posi(ii)
    dists = water.dist_via_oh_list(box, posi, oh_list)
    print(idx, np.min(dists), np.max(dists), np.average(dists))
          



#!/usr/bin/env python3

import numpy as np

def get_natoms(filename) :
    with open(filename, 'r') as fp :
        lines = fp.read().split('\n')
    for ii in lines :
        if 'atoms' in ii:
            natoms = int(ii.split()[0])
            return natoms
    raise RuntimeError("cannot find key word \'atoms\' in " + filename)

def _is_n_number(inputs, nn) :
    words = inputs.split()
    if len(words) != nn :
        return False, None
    res = []
    for ii in words :
        try :
            res.append(float(ii))
        except ValueError:
            return False, None
    return True, res

def get_thermo(filename) :
    with open(filename, 'r') as fp :
        fc = fp.read().split('\n')
    for sl in range(len(fc)) :
        if 'Step KinEng PotEng TotEng' in fc[sl] :
            break
    nwords = len(fc[sl+1].split())
    data = []
    for el in range(sl+1,len(fc)) :
        flag, res = _is_n_number(fc[el], nwords) 
        if not flag:
            break
        else :
            data.append(res)
    data = np.array(data)
    return data

def get_thermo_old(filename) :
    with open(filename, 'r') as fp :
        fc = fp.read().split('\n')
    for sl in range(len(fc)) :
        if 'Step KinEng PotEng TotEng' in fc[sl] :
            break
    for el in range(len(fc)) :
        if 'Loop time of' in fc[el] :
            break
    data = []
    for ii in range(sl+1, el) :
        data.append([float(jj) for jj in fc[ii].split()])
    data = np.array(data)
    return data

def get_last_dump(dump) :
    with open(dump, 'r') as fp :
        lines = fp.read().split('\n')
    lines = lines[:-1]
    step_idx = -1
    for idx,ii in enumerate(lines) :
        if 'ITEM: TIMESTEP' in ii :
            step_idx = idx
    if step_idx == -1 :
        raise RuntimeError("cannot find timestep in lammps dump, something wrong")    
    return '\n'.join(lines[step_idx:])



import os, textwrap
import numpy as np
import unittest
from context import dpti
# from potential_common import soft_param, soft_param_three_element, meam_model
from dpti.lib.lammps import get_natoms, get_thermo, get_last_dump
# from dpti.lib.dump import from_system_data
from numpy.testing import assert_almost_equal

class TestLibLammpsGetNatoms(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_normal(self):
        lmp_file = os.path.join('lammps_test_files', 'test_hti.lmp')
        natoms = get_natoms(lmp_file)
        self.assertEqual(144, natoms)

    def test_raise_err(self):
        lmp_file = os.path.join('lammps_test_files', 'test_hti.lmp.broken_test_case')
        with self.assertRaises(RuntimeError):
            get_natoms(lmp_file)

class TestGetThermo(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

        self.log_file = os.path.join('lammps_test_files', 
            'get_thermo.log.lammps')
        self.data_file = os.path.join('lammps_test_files', 
            'get_thermo.data')
        # dump fmt=%.12e

        # data1 = get_thermo(self.log_file)
        # np.savetxt(
        #     fname=self.data_file,
        #     X=data1,
        #     fmt='%.12e'
        # )

    def test_normal(self):
        data1 = get_thermo(self.log_file)
        data2 = np.loadtxt(self.data_file)
        assert_almost_equal(data1, data2, decimal=8)
    
    def test_raise_err(self):
        data1 = get_thermo(self.log_file)
        data2 = np.loadtxt(self.data_file)
        with self.assertRaises(AssertionError):
            assert_almost_equal(data1, data2, decimal=10)

class TestGetLastDump(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.last_dump_file = os.path.join(
            'lammps_test_files',
            'test_get_last_dump.dump.equi.lastdump'
        )
        self.dump_file = os.path.join(
            'lammps_test_files', 
            'test_get_last_dump.dump.equi'
        )

    def test_normal(self):
        with open(self.last_dump_file, 'r') as f:
            ret1_with_ending_newline = f.read()
        ret1 = ret1_with_ending_newline.strip('\n')
        ret2 = get_last_dump(self.dump_file)
        self.assertEqual(ret1, ret2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import os,sys
import numpy as np
lib_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append (lib_path)
import lmp as lmp


def posi_diff(box, r0, r1) :
    rbox = np.linalg.inv(box)
    rbox = rbox.T
    p0 = (np.dot(rbox, r0))
    p1 = (np.dot(rbox, r1))
    dp = p0 - p1
    shift = np.zeros(3)
    for dd in range(3) :
        if dp[dd] >= 0.5 : 
            dp[dd] -= 1
        elif dp[dd] < -0.5 :
            dp[dd] += 1
    dr = np.dot(box.T, dp)    
    return dr

def posi_shift(box, r0, r1) :
    rbox = np.linalg.inv(box)
    rbox = rbox.T
    p0 = (np.dot(rbox, r0))
    p1 = (np.dot(rbox, r1))
    dp = p0 - p1
    shift = np.zeros(3)
    for dd in range(3) :
        if dp[dd] >= 0.5 : 
            shift[dd] -= 1
        elif dp[dd] < -0.5 :
            shift[dd] += 1
    return shift

def compute_bonds(box, atype, posis, 
                  max_roh = 1.3, 
                  uniq_hbond = True):
    natoms = len(posis)
    bonds = []
    for ii in range(natoms) :
        bonds.append([])
    for ii in range(natoms) :
        if atype[ii] == 1 :
            for jj in range(natoms) :
                if atype[jj] == 2 :
                    dr = posi_diff(box, posis[ii], posis[jj])
                    if np.linalg.norm(dr) < max_roh :
                        bonds[ii].append(jj)
                        bonds[jj].append(ii)
    if uniq_hbond :
        for jj in range(natoms) :
            if atype[jj] == 2 :
                if len(bonds[jj]) > 1 :
                    orig_bonds = bonds[jj]
                    min_bd = 1e10
                    min_idx = -1
                    for ii in bonds[jj] :
                        dr = posi_diff(box, posis[ii], posis[jj])
                        drr = np.linalg.norm(dr)
                        # print(ii,jj, posis[ii], posis[jj], drr)
                        if drr < min_bd :
                            min_idx = ii
                            min_bd = drr
                    bonds[jj] = [min_idx]
                    orig_bonds.remove(min_idx)
                    # print(min_idx, orig_bonds, bonds[jj])
                    for ii in orig_bonds :
                        bonds[ii].remove(jj)
    return bonds


def add_bonds (lines_, max_roh = 1.3) :
    lines = lines_
    natoms = lmp.get_natoms_vec(lines)
    assert(len(natoms) == 2) 
    # type 1 == O, type 2 == H
    assert(natoms[0] == natoms[1] // 2) 
    
    atype = lmp.get_atype(lines)
    posis = lmp.get_posi(lines)
    bounds, tilt = lmp.get_lmpbox(lines)
    orig, box = lmp.lmpbox2box(bounds, tilt)
    bonds = compute_bonds(box, atype, posis, max_roh)

    # check water moles
    for ii in range(len(bonds)) :
        if atype[ii] == 1 :
            assert(len(bonds[ii]) == 2), 'ill defined O atom %d has H %s' % (ii, bonds[ii])
        elif atype[ii] == 2 :
            assert(len(bonds[ii]) == 1), 'ill defined H atom %d has O %s' % (ii, bonds[ii]) 

    # pbc posi
    for ii in range(sum(natoms)) :
        if atype[ii] == 1:
            j0idx = bonds[ii][0]
            shift0 = posi_shift(box, posis[j0idx], posis[ii])
            posis[j0idx] = posis[j0idx] + np.dot(box.T, shift0)
            j1idx = bonds[ii][1]
            shift1 = posi_shift(box, posis[j1idx], posis[ii])
            posis[j1idx] = posis[j1idx] + np.dot(box.T, shift1)
    for atoms_idx in range(len(lines)) :
        if 'Atoms' in lines[atoms_idx] :
            break
    for ii in range(sum(natoms)) :
        words = lines[atoms_idx + 2 +ii].split()
        words[2] = str(posis[ii][0])
        words[3] = str(posis[ii][1])
        words[4] = str(posis[ii][2])
        lines[atoms_idx+2+ii] = ' '.join(words)


    ret_bd = []
    idx = 1
    for ii in range(len(bonds)) :
        if atype[ii] == 1:
            ret_bd.append("%d 1 %d %d" % (idx, 1+ii, 1+bonds[ii][0]))
            idx += 1
            ret_bd.append("%d 1 %d %d" % (idx, 1+ii, 1+bonds[ii][1]))
            idx += 1

    ret_ang = []
    idx = 1
    for ii in range(len(bonds)) :
        if atype[ii] == 1:
            ret_ang.append("%d 1 %d %d %d" % (idx, 1+bonds[ii][0], 1+ii, 1+bonds[ii][1]))
            idx += 1

    lines.append('Bonds')
    lines.append('')
    lines += ret_bd
    lines.append('')
    lines.append('Angles')
    lines.append('')
    lines += ret_ang
    lines.append('')
    lines.append('')

    nbonds = len(ret_bd)
    nangles = len(ret_ang)
    for atoms_idx in range(len(lines)) :
        if 'atoms' in lines[atoms_idx] :
            break
    lines.insert(atoms_idx+1, '%d angles' % nangles)
    lines.insert(atoms_idx+1, '%d bonds' % nbonds)
    for atoms_types_idx in range(len(lines)) :
        if 'atom types' in lines[atoms_types_idx] :
            break
    lines.insert(atoms_types_idx+1, '%d angle types' % 1)
    lines.insert(atoms_types_idx+1, '%d bond types' % 1)

    for atoms_idx in range(len(lines)) :
        if 'Atoms' in lines[atoms_idx] :
            break
    mole_idx = np.zeros(sum(natoms), dtype = int)
    cc = 1
    for ii in range(sum(natoms)) :
        if atype[ii] == 1 :
            mole_idx[ii] = cc
            cc += 1
    for ii in range(sum(natoms)) :
        if atype[ii] == 2 :
            mole_idx[ii] = mole_idx[bonds[ii]]
    cc = 0
    for idx in range(atoms_idx+2, atoms_idx+2+sum(natoms)) :
        words = lines[idx].split()
        words.insert(1, '%d' % mole_idx[cc])
        lines[idx] = ' '.join(words)
        cc += 1

    return lines


def min_oo (box, atype, posis) :
    natoms = len(posis)
    min_dist = 1e10 * np.ones([natoms])
    for ii in range(natoms) :
        for jj in range(ii+1,natoms) :
            if atype[ii] == 1 and atype[jj] == 1 :
                dij = np.linalg.norm(posi_diff(box, posis[ii], posis[jj]))
                if dij < min_dist[ii] :
                    min_dist[ii] = dij
                if dij < min_dist[jj] :
                    min_dist[jj] = dij
    _min_dist = []
    for ii in min_dist :
        if ii != 1e10 :
            _min_dist.append(ii)
    return _min_dist


def min_ho (box, atype, posis) :
    natoms = len(posis)
    min_dist = 1e10 * np.ones([natoms])
    for ii in range(natoms) :
        for jj in range(natoms) :
            if atype[ii] == 2 and atype[jj] == 1 :
                dij = np.linalg.norm(posi_diff(box, posis[ii], posis[jj]))
                if dij < min_dist[ii] :
                    min_dist[ii] = dij
    _min_dist = []
    for ii in min_dist :
        if ii != 1e10 :
            _min_dist.append(ii)
    return _min_dist        


def min_oho (box, atype, posis) :
    natoms = len(posis)
    dist_oh = []
    dist_oh2 = []
    dist_oo = []
    for ii in range(natoms) :
        if atype[ii] == 1 :
            continue
        all_dist = []
        for jj in range(natoms) :
            if atype[jj] != 1 :
                continue
            dij = np.linalg.norm(posi_diff(box, posis[ii], posis[jj]))
            all_dist.append([dij, jj])
        all_dist.sort()
        doo = np.linalg.norm(posi_diff(box, 
                                       posis[all_dist[0][1]], 
                                       posis[all_dist[1][1]]
        ))
        dist_oh.append(all_dist[0][0])
        dist_oh2.append(all_dist[1][0])
        dist_oo.append(doo)
    # assume O-H..O
    # dist O-H, dist H..O, dist O-O
    return dist_oh, dist_oh2, dist_oo


def min_oh_list(box, atype, posis) :
    natoms = len(posis)
    list_oh = []
    for ii in range(natoms) :
        if atype[ii] == 1 :
            continue
        all_dist = []
        for jj in range(natoms) :
            if atype[jj] != 1 :
                continue
            dij = np.linalg.norm(posi_diff(box, posis[ii], posis[jj]))
            all_dist.append([dij, jj])
        all_dist.sort()
        doo = np.linalg.norm(posi_diff(box, 
                                       posis[all_dist[0][1]], 
                                       posis[all_dist[1][1]]
        ))
        list_oh.append([all_dist[0][1], ii])
    # assume O-H..O
    # dist O-H, dist H..O, dist O-O
    return list_oh


def dist_via_oh_list(box, posis, list_oh) :
    dist = []
    for ii in list_oh :
        dij = np.linalg.norm(posi_diff(box, posis[ii[0]], posis[ii[1]]))
        dist.append(dij)
    return dist


if __name__ == '__main__' :
    fname = 'conf.lmp'
    lines = open(fname).read().split('\n')
    # lines = add_bonds(lines)
    # print('\n'.join(ret_bd))
    # print('\n'.join(ret_ang))

    atype = lmp.get_atype(lines)
    posis = lmp.get_posi(lines)
    bounds, tilt = lmp.get_lmpbox(lines)
    orig, box = lmp.lmpbox2box(bounds, tilt)

    # md_oo = min_oo(box, atype, posis)
    # md_ho = min_ho(box, atype, posis)
    # print(np.average(md_oo), np.average(md_ho), np.min(md_ho), np.max(md_ho))

    moh, moh2, moo = min_oho(box, atype, posis)
    print(np.max(moh), np.average(moh2), np.average(moo))




import json, os, sys, glob

# from airflow.models import DAG
from datetime import datetime, timedelta
from airflow.decorators import dag, task
from airflow.operators.python import get_current_context
from airflow.operators.bash_operator import BashOperator
from airflow.operators.dummy_operator import DummyOperator
from airflow.operators.python_operator import PythonOperator
from airflow.utils.dates import days_ago

from airflow.exceptions import AirflowSkipException, AirflowFailException

# from dpdispatcher.lazy_local_context import LazyLocalContext
from dpdispatcher import Submission, Task, Resources, Machine
from dpti import equi, hti, hti_liq, ti
import subprocess as sp

def get_empty_submission(job_work_dir):
    context = get_current_context()
    dag_run = context['params']
    work_base_dir = dag_run['work_base_dir']

    with open(os.path.join(work_base_dir, 'machine.json'), 'r') as f:
        mdata = json.load(f)
    machine = Machine.load_from_dict(mdata['machine'])
    resources = Resources.load_from_dict(mdata['resources'])

    submission = Submission(
        work_base=job_work_dir, 
        resources=resources, 
        machine=machine, 
    )
    return submission

@task()
def all_start_check():
    context = get_current_context()
    print(context)
    print('debug77878', os.getcwd())

    dag_run = context['params']
    work_base_dir = dag_run['work_base_dir']
    target_temp = int(dag_run['target_temp'])
    target_pres = int(dag_run['target_pres'])
    conf_lmp = str(dag_run['conf_lmp'])
    ti_path = str(dag_run['ti_path'])
    ens = str(dag_run['ens'])
    if_liquid = dag_run['if_liquid']


    work_base_abs_dir = os.path.realpath(work_base_dir)

    dag_work_dirname=str(target_temp)+'K-'+str(target_pres)+'bar-'+str(conf_lmp)
    dag_work_dir=os.path.join(work_base_abs_dir, dag_work_dirname)

    assert os.path.isdir(work_base_dir) is True,  f'work_base_dir {work_base_dir} must exist '
    if os.path.isdir(dag_work_dir) is False:
        os.mkdir(dag_work_dir)
    else:
        pass

    conf_lmp_abs_path = os.path.join(work_base_abs_dir, conf_lmp)
    assert os.path.isfile(conf_lmp_abs_path) is True,  f'structure file {conf_lmp_abs_path} must exist'
    assert str(ti_path) in ["t", "p"], f'value for "path" must be "t" or "p" '

    start_info = dict(work_base_dir=work_base_dir, 
        target_temp=target_temp,
        target_pres=target_pres, 
        conf_lmp=conf_lmp, 
        ti_path=ti_path,
        ens=ens, 
        if_liquid=if_liquid, 
        work_base_abs_dir=work_base_abs_dir,
        dag_work_dir=dag_work_dir)
    print('start_info:', start_info)
    return start_info

@task()
def NPT_start(start_info):
    print(start_info)
    work_base_abs_dir = start_info['work_base_abs_dir']
    dag_work_dir = start_info['dag_work_dir']
    job_work_dir = os.path.join(dag_work_dir, 'NPT_sim', 'new_job')

    result_json_file = os.path.join(job_work_dir, 'result.json')
    if os.path.isfile(result_json_file):
        return job_work_dir
    with open(os.path.join(work_base_abs_dir, 'npt.json')) as f:
        npt_jdata = json.load(f)

    task_jdata = npt_jdata.copy()
    task_jdata['equi_conf'] = start_info['conf_lmp']
    task_jdata['temp'] = start_info['target_temp']
    task_jdata['pres'] = start_info['target_pres']
    task_jdata['ens'] = start_info['ens']
    print(task_jdata)
    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    equi.make_task(job_work_dir, task_jdata)
    os.chdir(cwd)
    return job_work_dir


@task(trigger_rule='none_failed_or_skipped')
def NPT_sim(job_work_dir):
    task = Task(command='lmp -i in.lammps', task_work_path='./', 
        forward_files=['in.lammps', '*lmp', 'graph.pb'], backward_files=['log.lammps', 'dump.equi']) # Files to transfer forward and backwards
    submission = get_empty_submission(job_work_dir)
    submission.register_task_list([task])
    submission.run_submission()
    return job_work_dir

@task(trigger_rule='none_failed')
def NPT_end(job_work_dir):
    result_file_path = os.path.join(job_work_dir, 'result.json')
    if os.path.isfile(result_file_path):
        info = json.load(open(result_file_path, 'r'))
    else:
        info = equi.post_task(job_work_dir)
    print(info)
    return info

@task()
def NVT_start(start_info, *, NPT_end_info):

    print('NPT_end_info', NPT_end_info)
    npt_dir = NPT_end_info['job_dir']
    print('debug', npt_dir)
    
    work_base_abs_dir = start_info.get('work_base_abs_dir')
    dag_work_dir = start_info.get('dag_work_dir')
    job_work_dir = os.path.join(dag_work_dir, 'NVT_sim', 'new_job')

    result_json_file = os.path.join(job_work_dir, 'result.json')
    if os.path.isfile(result_json_file):
        return job_work_dir

    with open(os.path.join(work_base_abs_dir, 'nvt.json')) as f:
        nvt_jdata = json.load(f)

    task_jdata = nvt_jdata.copy()
    task_jdata['temp'] = start_info['target_temp']
    task_jdata['pres'] = start_info['target_pres']
    task_jdata['ens'] = 'nvt' 
    print('debug', npt_dir)
    
    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    equi.make_task(job_work_dir, task_jdata, npt_dir=npt_dir)
    os.chdir(cwd)
    return job_work_dir

@task(trigger_rule='none_failed_or_skipped')
def NVT_sim(job_work_dir):
    submission = get_empty_submission(job_work_dir)
    task = Task(command='lmp -i in.lammps', task_work_path='./',
        forward_files=['in.lammps', '*lmp', 'graph.pb'], backward_files=['log.lammps'])
    submission.register_task_list([task])
    submission.run_submission()
    return job_work_dir

@task(trigger_rule='none_failed')
def NVT_end(job_work_dir):
    result_file_path = os.path.join(job_work_dir, 'result.json')
    if os.path.isfile(result_file_path):
        with open(result_file_path, 'r') as f:
            info = json.load(f)
    else:
        info = equi.post_task(job_work_dir)
    print(info)
    return info

@task()
def HTI_start(start_info, *, NVT_end_info={}):
    work_base_abs_dir = start_info['work_base_abs_dir']
    dag_work_dir = start_info['dag_work_dir']
    if_liquid = start_info['if_liquid']
    conf_lmp = NVT_end_info.get('out_lmp', None)

    job_work_dir = os.path.join(dag_work_dir, 'HTI_sim', 'new_job')
    if os.path.isfile(os.path.join(dag_work_dir, 'HTI_sim', 'new_job', 'result.json')):
        return job_work_dir

    if if_liquid:
        with open(os.path.join(work_base_abs_dir, 'hti.liquid.json')) as j:
            hti_jdata = json.load(j)
    else:
        with open(os.path.join(work_base_abs_dir, 'hti.json')) as j:
            hti_jdata = json.load(j)
    
    task_jdata = hti_jdata.copy()
    if conf_lmp is not None:
        task_jdata['equi_conf'] = conf_lmp
    else:
        task_jdata['equi_conf'] = start_info['conf_lmp']

    task_jdata['temp'] = start_info['target_temp']
    task_jdata['pres'] = start_info['target_pres']

    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    if if_liquid:
        hti_liq.make_tasks(iter_name=job_work_dir, jdata=task_jdata)
    else:
        hti.make_tasks(iter_name=job_work_dir, jdata=task_jdata, ref='einstein', switch='three-step')
    os.chdir(cwd)
    return job_work_dir

@task()
def HTI_sim(job_work_dir):
    task_abs_dir_list = glob.glob(os.path.join(job_work_dir, './*/task*'))
    task_dir_list = [os.path.relpath(ii, start=job_work_dir ) for ii in task_abs_dir_list]

    task_list = [ Task(command='lmp -i in.lammps', 
        task_work_path=ii, forward_files=['in.lammps', '*lmp', 'graph.pb'], backward_files=['log.lammps']) 
        for ii in task_dir_list ]
    submission = get_empty_submission(job_work_dir)
    # submission.forward_common_files = 
    submission.register_task_list(task_list=task_list)
    submission.run_submission()
    return job_work_dir

@task()
def HTI_end(job_work_dir,
    start_info,
    NPT_end_info={}
):
    if_liquid = start_info['if_liquid']
    manual_pv = NPT_end_info.get('pv')
    manual_pv_err = NPT_end_info.get('pv_err')

    result_file_path = os.path.join(job_work_dir, 'result.json')
    if os.path.isfile(result_file_path):
        info = json.load(open(result_file_path, 'r'))
        return info

    if if_liquid:
        info = hti_liq.compute_task(
            job_work_dir, 
            'gibbs', 
            manual_pv=manual_pv, 
            manual_pv_err=manual_pv_err
        )
    else:
        info = hti.compute_task(
            job_work_dir, 
            'gibbs', 
            manual_pv=manual_pv, 
            manual_pv_err=manual_pv_err
        )
    print(info)
    return info

@task()
def TI_start(start_info, *, HTI_end_info=None):
    work_base_abs_dir = start_info['work_base_abs_dir']
    dag_work_dir = start_info['dag_work_dir']

    ti_path = start_info['ti_path']
    conf_lmp = start_info['conf_lmp']

    if ti_path == 't':
        with open(os.path.join(work_base_abs_dir, 'ti.t.json')) as j:
            ti_jdata = json.load(j)
            task_jdata = ti_jdata.copy()
            task_jdata['pres'] = start_info['target_pres']
            job_dir = 'TI_t_sim'
    elif ti_path == 'p':
        with open(os.path.join(work_base_abs_dir, 'ti.p.json')) as j:
            ti_jdata = json.load(j)
            task_jdata = ti_jdata.copy()
            task_jdata['temp'] = start_info['target_temp']
            job_dir = 'TI_p_sim'
    else:
        raise RuntimeError(f'Error integration path. {ti_path}')

    job_work_dir = os.path.join(dag_work_dir, job_dir, 'new_job')
    task_jdata['ens'] = start_info['ens']
    task_jdata['equi_conf'] = conf_lmp

    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    ti.make_tasks(job_work_dir, task_jdata)
    os.chdir(cwd)
    return job_work_dir

@task()
def TI_sim(job_work_dir):
    task_abs_dir_list = glob.glob(os.path.join(job_work_dir, './task*'))
    task_dir_list = [os.path.relpath(ii, start=job_work_dir ) for ii in task_abs_dir_list]
    submission = get_empty_submission(job_work_dir)
    task_list = [ Task(command='lmp -i in.lammps', task_work_path=ii, 
         forward_files=['in.lammps', '*lmp', 'graph.pb'], backward_files=['log.lammps']) for ii in task_dir_list]
    submission.register_task_list(task_list=task_list)
    submission.run_submission()
    return job_work_dir

@task()
def TI_end(job_work_dir, start_info, HTI_end_info):
    Eo = HTI_end_info['e1']
    Eo_err = HTI_end_info['e1_err']

    ti_path = start_info['ti_path']
    if ti_path == 't':
        To = start_info['target_temp']
    if ti_path == 'p':
        To = start_info['target_pres']
    result_file_path = os.path.join(job_work_dir, 'result.json')
    if os.path.isfile(result_file_path):
        with open(result_file_path, 'r') as f:
            info = json.load(f)
    else:
        info = ti.compute_task(job_work_dir, inte_method='inte', Eo=Eo, Eo_err=Eo_err, To=To)
    print(info)
    return info

default_args = {
    'owner': 'fengbo',
    'retries': 0,
    'start_date': datetime(2021, 1, 1, 8, 00)
}

@dag(default_args=default_args, schedule_interval=None,)
def TI_taskflow():
    start_info = all_start_check()
    NPT_end_info = NPT_end(
        NPT_sim(NPT_start(start_info=start_info)))

    NVT_end_info = NVT_end(NVT_sim(NVT_start(
        start_info=start_info, NPT_end_info=NPT_end_info)))

    HTI_end_info = HTI_end(HTI_sim(HTI_start(
            start_info=start_info, 
            NVT_end_info=NVT_end_info)),
        start_info=start_info,
        NPT_end_info=NPT_end_info)

    TI_end_info = TI_end(TI_sim(TI_start(
        start_info=start_info, 
        HTI_end_info=HTI_end_info)), 
        start_info=start_info, 
        HTI_end_info=HTI_end_info)

    return TI_end_info

TI_dag = TI_taskflow()


